import sys

from threading import Thread
from collections import defaultdict, OrderedDict

from requests import Session

from .config import DEFAULT_ORIGIN
from .utils import get_url, get_csrf_token, get_oauth_token


class SpotifyLocal:
    """Controller for the local spotify web helper, throws events when the
    state of spotify changes.
    """

    def __init__(self):
        self._registered_events = defaultdict(OrderedDict)
        self._csrf_token = get_csrf_token()
        self._oauth_token = get_oauth_token()
        self._session = Session()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def _request(self, url, params={}):
        """Makes a request using the currently open session.

        :param url: A url fragment to use in the creation of the master url
        """
        r = self._session.get(url=url, params=params, headers=DEFAULT_ORIGIN)
        return r

    def close(self):
        self._session.close()

    def on(self, event):
        """Decorator function that adds function to callback list for event system
        There are three events you can subscribe too; **status_change**, **play_state_change**, **track_change**.

        :param event: name of the event you wish to register the function under, you can use multiple decorators
        """

        def _on(func):
            self.add_event_handler(event, func)
            return func

        return _on

    def add_event_handler(self, event, func):
        """Add function and event to Ordered Dict

        :param event: Name of the event you wish to register the function
        :param func: Function to register witht the associated event
        """
        self._registered_events[event][func] = func

    def emit(self, event, *args, **kwargs):
        """Send out an event and call it's associated functions

        :param event: Name of the event to trigger
        """
        for func in self._registered_events[event].values():
            func(*args, **kwargs)

    def remove_listener(self, event, func):
        """Remove an event listner

        :param event: Event you wish to remove the function from
        :param func: Function you wish to remove
        """
        self._registered_events[event].pop(func)

    def remove_all_listeners(self, event=None):
        """Remove all functions for all events, or one event if one is specifed.

        :param event: Optional event you wish to remove all functions from
        """
        if event is not None:
            self._registered_events[event] = OrderedDict()
        else:
            self._registered_events = defaultdict(OrderedDict)

    def listeners(self, event):
        """Return list of listners associated to a particular event

        :param event: Name of the event you wish to query
        """
        return list(self._registered_events[event].keys())

    @property
    def version(self):
        """Spotify version information"""
        url: str = get_url("/service/version.json")
        params = {"service": "remote"}
        r = self._request(url=url, params=params)
        return r.json()

    def get_current_status(self):
        """Returns the current state of the local spotify client"""
        url = get_url("/remote/status.json")
        params = {"oauth": self._oauth_token, "csrf": self._csrf_token}
        r = self._request(url=url, params=params)
        return r.json()

    def pause(self, pause=True):
        """Pauses the spotify player

        :param pause: boolean value to choose the pause/play state
        """
        url: str = get_url("/remote/pause.json")
        params = {
            "oauth": self._oauth_token,
            "csrf": self._csrf_token,
            "pause": "true" if pause else "false",
        }
        self._request(url=url, params=params)

    def unpause(self):
        """Unpauses the player by calling pause()"""
        self.pause(pause=False)

    def playURI(self, uri):
        """Play a Spotify uri, for example spotify:track:5Yn8WCB4Dqm8snemB5Mu4K

        :param uri: Playlist, Artist, Album, or Song Uri
        """
        url: str = get_url("/remote/play.json")
        params = {
            "oauth": self._oauth_token,
            "csrf": self._csrf_token,
            "uri": uri,
            "context": uri,
        }
        r = self._request(url=url, params=params)
        return r.json()

    @staticmethod
    def skip():
        """Skips the current song"""
        if sys.platform == "darwin":
            keyboard.send("KEYTYPE_SKIP")
        else:
            keyboard.send("next track")

    @staticmethod
    def previous():
        """Goes to the beginning of the track, or if called twice goes to the previous track."""
        if sys.platform == "darwin":
            keyboard.send("KEYTYPE_PREVIOUS")
        else:
            keyboard.send("previous track")

    def listen(self, wait=60, blocking=True) -> None:
        """Listen for events and call any associated callbacks when there is an event.
        There are three events you can subscribe too; **status_change**, **play_state_change**, **track_change**

        :param wait: how long to wait for a response before starting a new connection
        :param blocking: if listen should block the current process or not
        """
        url = get_url("/remote/status.json")

        def listen_for_status_change():
            params = {"oauth": self._oauth_token, "csrf": self._csrf_token}
            r = self._request(url=url, params=params)
            old = r.json()
            self.emit("status_change", old)
            params = {
                "oauth": self._oauth_token,
                "csrf": self._csrf_token,
                "returnon": "play,pause,error,ap",
                "returnafter": wait,
            }
            while True:
                r = self._request(url=url, params=params)
                new = r.json()

                if new != old:
                    self.emit("status_change", new)

                if new["playing"] != old["playing"]:
                    self.emit("play_state_change", new)

                if (
                    new["track"]["track_resource"]["uri"]
                    != old["track"]["track_resource"]["uri"]
                ):
                    self.emit("track_change", new)

                old = new

        if blocking:
            listen_for_status_change()
        else:
            thread = Thread(target=listen_for_status_change)
            thread.daemon = True
            thread.start()



import sys

try:
    assert sys.version_info.major == 3
    assert sys.version_info.minor > 2
except AssertionError:
    raise RuntimeError("Spotify-Local requires Python 3.6+!")

from .core import SpotifyLocal

__version__ = "0.3.0"




from random import choices
from string import ascii_lowercase

from requests import Response, session, Session

from .config import DEFAULT_PORT, DEFAULT_ORIGIN

s: Session = session()


def get_url(url: str) -> str:
    """Ranomdly generates a url for use in requests.
    Generates a hostname with the port and the provided suffix url provided
    :param url: A url fragment to use in the creation of the master url
    """
    sub = "{0}.spotilocal.com".format("".join(choices(ascii_lowercase, k=10)))
    return "http://{0}:{1}{2}".format(sub, DEFAULT_PORT, url)


def get_oauth_token() -> str:
    """Retrieve a simple OAuth Token for use with the local http client."""
    url: str = "{0}/token".format(DEFAULT_ORIGIN["Origin"])
    r: Response = s.get(url=url)
    return r.json()["t"]


def get_csrf_token() -> str:
    """Retrieve a simple csrf token for to prevent cross site request forgery."""
    url: str = get_url("/simplecsrf/token.json")
    r: Response = s.get(url=url, headers=DEFAULT_ORIGIN)
    return r.json()["token"]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Note: To use the 'upload' functionality of this file, you must:
#   $ pip install twine

import io
import os
import sys
import re
from shutil import rmtree

from setuptools import setup, Command, find_packages


here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.rst' is present in your MANIFEST.in file!
with io.open(os.path.join(here, "README.rst"), encoding="utf-8") as f:
    long_description = "\n" + f.read()


def find_version(*file_paths):
    with io.open(os.path.join(here, *file_paths), encoding="utf-8") as f:
        version_file = f.read()
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


# Package meta-data.
NAME = "spotify-local"
DESCRIPTION = "Communicate with the Spotify's web helper process to control some basic aspects of spotify"
URL = "https://github.com/erinxocon/spotify-local-control"
EMAIL = "erinocon5@gmail.com"
AUTHOR = "Erin O'Connell"
VERSION = find_version("src", "spotify_local", "__init__.py")

# What packages are required for this module to be executed?
REQUIRED = [
    "requests",
    "keyboard",
    "pyobjc-framework-Quartz; sys.platform == 'darwin'",
    "aiohttp",
]


class UploadCommand(Command):
    """Support setup.py upload."""

    description = "Build and publish the package."
    user_options = []  # type: ignore

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print("\033[1m{0}\033[0m".format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status("Removing previous builds…")
            rmtree(os.path.join(here, "dist"))
        except OSError:
            pass

        self.status("Building Source and Wheel (universal) distribution…")
        os.system("{0} setup.py sdist bdist_wheel --universal".format(sys.executable))

        self.status("Uploading the package to PyPi via Twine…")
        os.system("twine upload dist/*")

        self.status("Publishing git tags…")
        os.system("git tag v{0}".format(VERSION))
        os.system("git push --tags")

        sys.exit()


# Where the magic happens:
setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=long_description,
    author=AUTHOR,
    author_email=EMAIL,
    url=URL,
    python_requires=">=3.6.0",
    package_dir={"": "src"},
    packages=find_packages(where="src", exclude=["docs", "tests*"]),
    install_requires=REQUIRED,
    include_package_data=True,
    license="MIT",
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
    # $ setup.py publish support.
    cmdclass={"upload": UploadCommand},
)
